# --------------------------------------------------------------------
# Copyright (c) iEXBase. All rights reserved.
# Licensed under the MIT License.
# See License.txt in the project root for license information.
# --------------------------------------------------------------------

# Here we specify default values for the different needed urls.
# They are verified. Don't change this unless you know what you're doing.
DEFAULT_NODES = {
    'full_node': 'https://apilg.lgcyscan.network/',
    'solidity_node': 'https://apilg.lgcyscan.network/',
    'event_server': 'https://apilg.lgcyscan.network/'
}
